import re
import numbers
from .utility import convert_units
import numpy as np

class Gas(object):
    """ PhreeqPy Solution Class """
    def __init__(self, phreeqpython, number):
        self.pp = phreeqpython
        self.number = number

    def copy(self):
        """ Create a new copy, with unique solution number, from this solution """
        return self.pp.copy_gas(self.number)

    def forget(self):
        """ Create a new copy, with unique solution number, from this solution """
        return self.pp.remove_gases([self.number])

    @property
    def pressure(self):
        return self.pp.ip.get_gas_pressure(self.number)
    @property
    def volume(self):
        return self.pp.ip.get_gas_volume(self.number) 
    @property
    def total_moles(self):
        return self.pp.ip.get_gas_total_moles(self.number) 

    @property
    def components(self):
        return self.pp.ip.get_gas_components_moles(self.number) 
    @property
    def fractions(self):
        return self.pp.ip.get_gas_components_fractions(self.number) 
    @property
    def partial_pressures(self):
        return self.pp.ip.get_gas_components_pressures(self.number) 

    @property
    def dry_fractions(self):
        try:
            total = self.total_moles - self.components['H2O(g)']
        except:
            total = self.total_moles

        return {name: value/total for (name, value) in self.components.items()}



# -*- coding: utf-8 -*-*

"""Access PHREEQC-DLL via ctypes.

This is exchangeable with the COM interface.
"""

import ctypes
import os
import sys

if sys.version_info[0] == 2:
    #pylint: disable-msg=W0622
    def bytes(str_, encoding): #pylint: disable-msg=W0613
        """Compatibilty function for Python 3.
        """
        return str_
    range = xrange #pylint: disable-msg=C0103
    #pylint: enable-msg=W0622


class VIPhreeqc(object):
    """Wrapper for the VIPhreeqc DLL.
    """
    # All methods in `method_mapping` are added dynamically.
    # Therefore, pylint complains.
    #pylint: disable-msg=E1101
    def __init__(self, dll_path=None):
        """Connect to DLL and create IPhreeqc.

        The optional `dll_path` takes a path to the IPhreeqc shared library.
        If not provided it tries to select an appropriate library.
        Make sure you have the right library for your operating system.
        You may download one from here:
        ftp://brrftp.cr.usgs.gov/pub/charlton/iphreeqc/

        See the PhreeqPy documentation for help on compiling a IPhreeqc shared
        library.
        """
        if not dll_path:
            if sys.platform == 'win32':
                dll_name = './lib/VIPhreeqc.dll'
            elif 'linux' in sys.platform:
                dll_name = './lib/viphreeqc.so'
            elif sys.platform == 'darwin':
                dll_name = './lib/viphreeqc.dylib'
            else:
                msg = 'Platform %s is not supported.' % sys.platform
                raise NotImplementedError(msg)
            dll_path = os.path.join(os.path.dirname(__file__), dll_name)
        phreeqc = ctypes.cdll.LoadLibrary(dll_path)
        self.debug = False
        self.dll = phreeqc
        c_int = ctypes.c_int
        method_mapping = [('_accumulate_line', phreeqc.AccumulateLine,
                           [c_int, ctypes.c_char_p], c_int),
                          ('_add_error', phreeqc.AddError,
                           [c_int, ctypes.c_char_p], c_int),
                          ('_add_error', phreeqc.AddWarning,
                           [c_int, ctypes.c_char_p], c_int),
                          ('_clear_accumlated_lines',
                           phreeqc.ClearAccumulatedLines, [c_int], c_int),
                          ('_create_iphreeqc', phreeqc.CreateIPhreeqc,
                           [ctypes.c_void_p], c_int),
                          ('_destroy_iphreeqc', phreeqc.DestroyIPhreeqc,
                           [c_int], c_int),
                          ('_get_component', phreeqc.GetComponent,
                           [c_int, c_int], ctypes.c_char_p),
                          ('_get_component_count', phreeqc.GetComponentCount,
                           [c_int], c_int),
                          ('_get_dump_string', phreeqc.GetDumpString,
                           [c_int], ctypes.c_char_p),
                          ('_set_dump_string_on', phreeqc.SetDumpStringOn,
                           [c_int], c_int),
                          ('_get_error_string', phreeqc.GetErrorString,
                           [c_int], ctypes.c_char_p),
                          ('_get_selected_output_column_count',
                           phreeqc.GetSelectedOutputColumnCount, [c_int],
                           c_int),
                          ('_get_selected_output_row_count',
                           phreeqc.GetSelectedOutputRowCount, [c_int], c_int),
                          ('_get_value', phreeqc.GetSelectedOutputValue,
                           [c_int, c_int, c_int, ctypes.POINTER(VAR)], c_int),
                          ('_load_database', phreeqc.LoadDatabase,
                           [c_int, ctypes.c_char_p], c_int),
                          ('_load_database_string', phreeqc.LoadDatabaseString,
                           [c_int, ctypes.c_char_p], c_int),
                          ('_run_string', phreeqc.RunString,
                           [c_int, ctypes.c_char_p], c_int),
                          ('_set_selected_output_file_on',
                           phreeqc.SetSelectedOutputFileOn, [c_int, c_int],
                           c_int),
                          # VIPHREEQC Additions:
                          # gas
                          ('_get_gas_volume', phreeqc.GetGasVolume,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_gas_pressure', phreeqc.GetGasPressure,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_gas_total_moles', phreeqc.GetGasTotalMoles,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_gas_components', phreeqc.GetGasComponents,
                           [c_int, c_int], ctypes.c_char_p),
                          ('_get_gas_component_moles', phreeqc.GetGasComponentMoles,
                           [c_int, c_int, ctypes.c_char_p], ctypes.c_double),
                          # solution
                          ('_get_ph', phreeqc.GetPH,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_pe', phreeqc.GetPe,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_sc', phreeqc.GetSC,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_mu', phreeqc.GetMu,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_temperature', phreeqc.GetTemperature,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_mass', phreeqc.GetMass,
                           [c_int, c_int], ctypes.c_double),
                          ('_get_total', phreeqc.GetTotal,
                           [c_int, c_int, ctypes.c_char_p], ctypes.c_double),
                          ('_get_total_element', phreeqc.GetTotalElement,
                           [c_int, c_int, ctypes.c_char_p], ctypes.c_double),
                          ('_get_moles', phreeqc.GetMoles,
                           [c_int, c_int, ctypes.c_char_p], ctypes.c_double),
                          ('_get_activity', phreeqc.GetActivity,
                           [c_int, c_int, ctypes.c_char_p], ctypes.c_double),
                          ('_get_molality', phreeqc.GetMolality,
                           [c_int, c_int, ctypes.c_char_p], ctypes.c_double),
                          ('_get_species', phreeqc.GetSpecies,
                           [c_int, c_int], ctypes.c_char_p),
                          ('_get_species_masters', phreeqc.GetSpeciesMasters,
                           [c_int, c_int], ctypes.c_char_p),
                          ('_get_phases', phreeqc.GetPhases,
                           [c_int, c_int], ctypes.c_char_p),
                          ('_get_elements', phreeqc.GetElements,
                           [c_int, c_int], ctypes.c_char_p),
                          ('_get_si', phreeqc.GetSI,
                           [c_int, c_int, ctypes.c_char_p], ctypes.c_double),
                          ('_get_solution_list', phreeqc.GetSolutionList,
                           [c_int], ctypes.c_char_p)
                         ]
        for name, com_obj, argtypes, restype in method_mapping:
            com_obj.argtypes = argtypes
            com_obj.restype = restype
            setattr(self, name, com_obj)
        self.var = VAR()
        self.phc_error_count = 0
        self.phc_warning_count = 0
        self.phc_database_error_count = 0
        self.id_ = self.create_iphreeqc()

    @staticmethod
    def raise_ipq_error(error_code):
        """There was an error, raise an exception.
        """
        error_types = {0: 'ok', -1: 'out of memory', -2: 'bad value',
                       -3: 'invalid argument type', -4: 'invalid row',
                       -5: 'invalid_column', -6: 'invalid instance id'}
        error_type = error_types[error_code]
        if error_type:
            raise PhreeqcException(error_type)

    def raise_string_error(self, errors):
        """Raise an exception with message from IPhreeqc error.
        """
        if errors > 1:
            msg = '%s errors occured.\n' % errors
        elif errors == 1:
            msg = 'An error occured.\n'
        else:
            msg = 'Wrong error number.'
        raise Exception(msg + self.get_error_string())

    def accumulate_line(self, line):
        """Put line in input buffer.
        """
        errors = self._accumulate_line(self.id_, bytes(line, 'utf-8'))
        if errors != 0:
            self.raise_string_error(errors)

    def add_error(self, phc_error_msg):
        """Add an error message to Phreeqc.
        """
        errors = self._add_error(self.id_, bytes(phc_error_msg, 'utf-8'))
        if errors < 0:
            self.raise_string_error(errors)
        else:
            self.phc_error_count = errors

    def add_warning(self, phc_warn_msg):
        """Add an warning message to Phreeqc.
        """
        errors = self._add_warning(self.id_, bytes(phc_warn_msg, 'utf-8'))
        if errors < 0:
            self.raise_string_error(errors)
        else:
            self.phc_warning_count = errors

    def clear_accumlated_lines(self):
        """Clear the input buffer.
        """
        errors = self._clear_accumlated_lines(self.id_)
        if errors != 0:
            self.raise_string_error(errors)

    @property
    def column_count(self):
        """Get number of columns in selected output.
        """
        return self._get_selected_output_column_count(self.id_)

    def create_iphreeqc(self):
        """Create a IPhreeqc object.
        """
        error_code = self._create_iphreeqc(ctypes.c_void_p())
        if error_code < 0:
            self.raise_ipq_error(error_code)
        id_ = error_code
        return id_

    def destroy_iphreeqc(self):
        """Delete the current instance of IPhreeqc.
        """
        error_code = self._destroy_iphreeqc(self.id_)
        if error_code < 0:
            self.raise_ipq_error(error_code)

    def get_component(self, index):
        """Get one component.
        """
        component = self._get_component(self.id_, index).decode('utf-8')
        if not component:
            raise IndexError('No component for index %s' % index)
        return component

    @property
    def component_count(self):
        """Return the number of components.
        """
        return self._get_component_count(self.id_)

    def get_component_list(self):
        """Return all component names.
        """
        get_component = self.get_component
        return [get_component(index) for index in range(self.component_count)]

    def get_error_string(self):
        """Retrieves the error messages.
        """
        return self._get_error_string(self.id_).decode('utf-8')

    # Vitens VIPHREEQC Extensions

    # surface
    def get_thickness(self, surface):
        return self._get_surface_thickness(self.id_, surface)

    # gas
    def get_gas_volume(self, gas):
        return self._get_gas_volume(self.id_, gas)
    # gas
    def get_gas_pressure(self, gas):
        return self._get_gas_pressure(self.id_, gas)
    # gas
    def get_gas_total_moles(self, gas):
        return self._get_gas_total_moles(self.id_, gas)
    def get_gas_components(self, gas):
        return self._get_gas_components(self.id_, gas).decode('utf-8').split(",")
    def get_gas_component_moles(self, gas, component):
        return self._get_gas_component_moles(self.id_, gas, bytes(component, 'utf-8'))

    def get_gas_components_moles(self, gas):
        component_list = self.get_gas_components(gas)
        component_moles = {}
        for component in component_list:
            component_moles[component] = self.get_gas_component_moles(gas, component)

        return component_moles

    def get_gas_components_fractions(self, gas):
        total_moles = self.get_gas_total_moles(gas)

        return {name: value/total_moles for (name, value) in self.get_gas_components_moles(gas).items()}

    def get_gas_components_pressures(self, gas):
        total_moles = self.get_gas_total_moles(gas)
        total_pressure = self.get_gas_pressure(gas)

        return {name: value/total_moles * total_pressure for (name, value) in self.get_gas_components_moles(gas).items()}




    # solution
    def get_ph(self, solution):
        return self._get_ph(self.id_, solution)
    def get_pe(self, solution):
        return self._get_pe(self.id_, solution)
    def get_sc(self, solution):
        return self._get_sc(self.id_, solution)
    def get_mu(self, solution):
        return self._get_mu(self.id_, solution)
    def get_temperature(self, solution):
        return self._get_temperature(self.id_, solution)
    def get_mass(self, solution):
        return self._get_mass(self.id_, solution)
    def get_total(self, solution, element):
        return self._get_total(self.id_, solution, bytes(element, 'utf-8'))
    def get_total_element(self, solution, element):
        return self._get_total_element(self.id_, solution, bytes(element, 'utf-8'))
    def get_moles(self, solution, species):
        return self._get_moles(self.id_, solution, bytes(species, 'utf-8'))
    def get_activity(self, solution, species):
        return self._get_activity(self.id_, solution, bytes(species, 'utf-8'))
    def get_molality(self, solution, species):
        return(self._get_molality(self.id_, solution, bytes(species, 'utf-8')))
    def get_species_moles(self, solution):
        """ Returns a list of species and their molarity """
        species_list = self.get_species(solution)
        species_moles = {}
        for species in species_list:
            species_moles[species] = self.get_moles(solution, species)
        return species_moles
    def get_species_molalities(self, solution):
        """ Returns a list of species and their molality """
        species_list = self.get_species(solution)
        species_moles = {}
        for species in species_list:
            species_moles[species] = self.get_molality(solution, species)
        return species_moles
    def get_species_activities(self, solution):
        """ Returns a list of species and their molality """
        species_list = self.get_species(solution)
        species_moles = {}
        for species in species_list:
            species_moles[species] = self.get_activity(solution, species)
        return species_moles

    def get_species_masters(self, solution):
        """ Returns a dict of species and their masters """
        species_list = self._get_species_masters(self.id_, solution).decode('utf-8').split(";")
        species_dict = {}
        for specie in species_list:
            species_dict[specie.split(":")[0]] = specie.split(":")[1].split(",")[:-1]
        return species_dict

    def get_masters_species(self, solution):
        masters_list = {}
        species_dict = self.get_species_masters(solution)

        for specie, masters in species_dict.items():
            for master in masters:
                masters_list.setdefault(master,[]).append(specie)
            
        return masters_list

    def get_solution_list(self):
        solution_list = self._get_solution_list(self.id_).decode('utf-8').split(",")
        return list(map(int, solution_list))
    def get_species(self, solution):
        return self._get_species(self.id_, solution).decode('utf-8').split(",")
    def get_si(self, solution, phase):
        return self._get_si(self.id_, solution, bytes(phase, 'utf-8'))
    def get_phases(self, solution):
        # no idea why this is necessary.. it won't work otherwise
        return self.dll.GetPhases(self.id_, solution).decode('utf-8').split(",")
    def get_phases_si(self, solution):
        """ Returns a list of phases and their solubility index """
        phases = self.get_phases(solution)
        phases_si = {}
        for phase in phases:
            phases_si[phase] = self.get_si(solution, phase)
        return phases_si
    def get_elements(self, solution):
        return self._get_elements(self.id_, solution).decode('utf-8').split(",")

    def get_elements_totals(self, solution):
        """ Returns a list of elements and their totals """
        elements = self.get_elements(solution)
        element_total = {}
        for element in elements:
            element_total[element] = self.get_total(solution, element)
        return element_total

    def set_dump_string_on(self):
        return self._set_dump_string_on(self.id_,1)
    def set_dump_string_off(self):
        return self._set_dump_string_on(self.id_,0)

    def get_dump_string(self):
        return self._get_dump_string(self.id_)
    # END Vitens extensions

    def get_selected_output_value(self, row, col):
        """Get one value from selected output at given row and column.
        """
        error_code = self._get_value(self.id_, row, col, self.var)
        if error_code != 0:
            self.raise_ipq_error(error_code)
        type_ = self.var.type
        value = self.var.value
        if type_ == 3:
            val = value.double_value
        elif type_ == 2:
            val = value.long_value
        elif type_ == 4:
            val = value.string_value.decode('utf-8')
        elif type_ == 0:
            val = None
        if type_ == 1:
            self.raise_error(value.error_code)
        return val

    def get_selected_output_array(self):
        """Get all values from selected output.
        """
        nrows = self.row_count
        ncols = self.column_count
        results = []
        for row in range(nrows):
            result_row = []
            for col in range(ncols):
                result_row.append(self.get_selected_output_value(row, col))
            results.append(result_row)
        return results

    def get_selected_output_row(self, row):
        """Get all values for one from selected output.
        """
        if row < 0:
            row = self.row_count + row
        ncols = self.column_count
        results = []
        for col in range(ncols):
            results.append(self.get_selected_output_value(row, col))
        return results

    def get_selected_output_column(self, col):
        """Get all values for one column from selected output.
        """
        if col < 0:
            col = self.column_count + col
        nrows = self.row_count
        results = []
        for row in range(nrows):
            results.append(self.get_selected_output_value(row, col))
        return results

    def set_selected_output_file_off(self):
        """Turn on writing to selected output file.
        """
        self._set_selected_output_file_on(self.id_, 0)

    def set_selected_output_file_on(self):
        """Turn on writing to selected output file.
        """
        self._set_selected_output_file_on(self.id_, 1)

    def load_database(self, database_name):
        """Load a database with given file_name.
        """
        self.phc_database_error_count = self._load_database(
            self.id_, bytes(database_name, 'utf-8'))

    def load_database_string(self, input_string):
        """Load a datbase from a string.
        """
        self.phc_database_error_count = self._load_database_string(
            self.id_, ctypes.c_char_p(bytes(input_string, 'utf-8')))

    @property
    def row_count(self):
        """Get number of rows in selected output.
        """
        return self._get_selected_output_row_count(self.id_)

    def run_string(self, cmd_string):
        """Run PHREEQC input from string.
        """
        #print(cmd_string)
        if self.debug:
            print(cmd_string)

        errors = self._run_string(self.id_,
                                  ctypes.c_char_p(bytes(cmd_string, 'utf-8')))
        if errors != 0:
            self.raise_string_error(errors)


class VARUNION(ctypes.Union):
    # pylint: disable-msg=R0903
    # no methods
    """Union with types.

    See Var.h in PHREEQC source.
    """
    _fields_ = [('long_value', ctypes.c_long),
                ('double_value', ctypes.c_double),
                ('string_value', ctypes.c_char_p),
                ('error_code', ctypes.c_int)]


class VAR(ctypes.Structure):
    # pylint: disable-msg=R0903
    # no methods
    """Struct with data type and data values.

    See Var.h in PHREEQC source.
    """
    _fields_ = [('type', ctypes.c_int),
                ('value', VARUNION)]

class PhreeqcException(Exception):
    """Error in Phreeqc call.
    """
    pass



from periodictable import formula as chemform

def convert_units(formula, amount, from_units='mol', to_units='mmol'):
    if formula == 'F' :
        formula = 'Ni'
    if from_units == to_units:
        return amount

    if from_units == 'mol':
        if to_units == 'mmol':
            return amount*1e3
        if to_units == 'mg':
            return chemform(formula).mass * amount * 1e3
        if to_units == 'ug':
            return chemform(formula).mass * amount * 1e6

    if from_units == 'mmol':
        if to_units == 'mol':
            return amount * 1e-3
        if to_units == 'mg':
            return chemform(formula).mass * amount
        if to_units == 'ug':
            return chemform(formula).mass * amount * 1e3

    if from_units == 'mg':
        if to_units == 'mol':
            return amount / chemform(formula).mass * 1e-3
        if to_units == 'mmol':
            return amount / chemform(formula).mass
        if to_units == 'ug':
            return amount * 1e3

    if from_units == 'ug': #micrograms
        if to_units == 'mol':
            return amount / chemform(formula).mass * 1e-6
        if to_units == 'mmol':
            return amount / chemform(formula).mass * 1e-3
        if to_units == 'mg':
            return amount * 1e-3




import re
import numbers
from .utility import convert_units
from scipy.integrate import odeint
import numpy as np

class Solution(object):
    """ PhreeqPy Solution Class """

    def __init__(self, phreeqpython, number):
        self.pp = phreeqpython
        self.factor = 1
        self.number = number

    def copy(self):
        """ Create a new copy, with unique solution number, from this solution """
        return self.pp.copy_solution(self.number)

    def add(self, element, amount, units='mmol'):
        """ Add a chemical to the solution """
        # convert to mol
        amount = convert_units(element, amount, units, 'mol')
        self.pp.change_solution(self.number, {element:amount})
        return self

    def remove(self, element, amount, units='mmol'):
        """ Remove a chemical from the solution """
        amount = -convert_units(element, amount, units, 'mol')
        self.pp.change_solution(self.number, {element:amount})
        return self


    def remove_fraction(self, species, fraction):
        """ Remove a fraction of a chemical from the solution """
        current = self.total(species)
        to_remove = current * fraction
        self.remove(species, to_remove)
        return self

    def interact(self, gas):
        self.pp.interact_solution_gas(self.number, gas.number)
        return self


    def equalize(self, phases, to_si=[0], in_phase=[10], with_chemical=[None]):
        """ equalize one or more phases with the solution """
        self.pp.equalize_solution(self.number, phases, to_si, in_phase, with_chemical)

        return self

    def saturate(self, phase, to_si=0, in_phase=10):
        """ Saturate a single phase to the given SI.
        This function can saturate one or multiple phases
        through dissolution. The maximum amount that can be dissolved is given by in_phase
        """
        if(self.si(phase) < 0):
            self.pp.equalize_solution(self.number, phase, to_si, in_phase)

        return self

    # this function can only precipitate
    def desaturate(self, phase, to_si=0):
        """ Desaturate a phase to the given SI.
        This function can only desaturate a phase through precipitation
        """
        self.pp.equalize_solution(self.number, phase, to_si, 0)

        return self

    # change the ph
    def change_ph(self, to_pH, with_chemical=None):
        """ Change the pH of a solution by dosing either HCl and NaOH, or a user supplied acid or base """
        # default to NaOH and HCl
        if not with_chemical:
            if to_pH < self.pH:
                # dose HCl to lower pH
                self.pp.equalize_solution(self.number, "Fix_pH", -to_pH, 10, "HCl")
            else:
                # dose NaOH to raise pH
                self.pp.equalize_solution(self.number, "Fix_pH", -to_pH, 10, "NaOH")
        else:
            self.pp.equalize_solution(self.number, "Fix_pH", -to_pH, 10, with_chemical)
        return self

    def change_temperature(self, to_temperature):
        """ Change the temperature of a solution """
        self.pp.change_solution_temperature(self.number, to_temperature)
        return self

    def total(self, element, units='mmol'):
        """ Returns to total of any given species or element (SLOW!) """
        total = 0
        regexp = "(^|[^A-Z])"+element
        for species, amount in self.species.items():
            if re.search(regexp, species):
                total += convert_units(element, amount, to_units=units)
        return total

    def total_activity(self, element, units='mmol'):
        """ Returns to total of any given species or element (SLOW!) """
        total = 0
        regexp = "(^|[^A-Z])"+element
        for species, amount in self.species_activities.items():
            if re.search(regexp, species):
                total += convert_units(element, amount, to_units=units)
        return total

    def total_element(self, element, units='mmol'):
        """ Returns to total any given element (FAST!) """
        return convert_units(element, self.pp.ip.get_total_element(self.number, element), 'mol', units)

    def activity(self, species, units='mmol'):
        """ Returns the activity of a single species """
        return convert_units(species, self.pp.ip.get_activity(self.number, species), 'mol', units)

    def moles(self, species, units='mmol'):
        """ Returns the moles of a single species """
        return convert_units(species, self.pp.ip.get_moles(self.number, species), 'mol', units)

    def molality(self, species, units='mmol'):
        """ Returns the molality of a single species """
        return convert_units(species, self.pp.ip.get_molality(self.number, species), 'mol', units)

    def si(self, phase):
        """ return the SI of a certain phase """
        return self.pp.ip.get_si(self.number, phase)

    def sr(self, phase):
        """ return the SI of a certain phase """
        return 10**self.pp.ip.get_si(self.number, phase)

    def forget(self):
        """ remove this solution from VIPhreeqc memory """
        self.pp.remove_solutions([self.number])

    def kinetics(self, element, rate_function, time, m0=0, args=()):

        def calc_rate(y, t, m0, *args):
            temp = self.copy()
            temp.add(element, y[0])
            rate = rate_function(temp, y[0], m0, *args)
            temp.forget()
            return rate

        y = odeint(calc_rate, 0, time, args=(m0,)+args)

        y = np.insert(np.diff(y[:,0]), 0, 0)

        for i in range(len(time)):
            t = time[i]
            self.add(element, y[i])
            yield(t, self)

    # Magic functions
    def __add__(self, other):
        """ add two solutions """
        if not isinstance(other,Solution):
            raise TypeError("Invalid operation, only addition of two solutions is allowed")
        mixture= {self:self.factor, other:other.factor}
        #print mixture
        mixture = self.pp.mix_solutions({self:self.factor,other:other.factor})
        # reset factors to 1
        self.factor = 1
        other.factor = 1

        return mixture

    def __truediv__(self, other):
        """ Python 3 support """
        return self.__div__(other)

    def __div__(self, other):
        """ set devision factor """
        if not isinstance(other,numbers.Real):
            raise TypeError("Invalid operation, only division by a number is allowed")
        self.factor = 1/float(other)
        return self

    def __mul__(self, other):
        """ set multiplication factor """
        if not isinstance(other,numbers.Real):
            raise TypeError("Invalid operation, only division by a number is allowed")
        self.factor = float(other)
        return self

    # Accessor methods
    @property
    def I(self):
        """ Solution ionic strength """
        return self.pp.ip.get_mu(self.number)
    def mu(self):
        """ Solution ionic strength """
        return self.I
    @property
    def pH(self):
        """ Solution pH """
        return self.pp.ip.get_ph(self.number)
    @property
    def sc(self):
        return self.pp.ip.get_sc(self.number)
    @property
    def temperature(self):
        return self.pp.ip.get_temperature(self.number)
    @property
    def mass(self):
        return self.pp.ip.get_mass(self.number)
    @property
    def pe(self):
        return self.pp.ip.get_pe(self.number)
    @property
    def phases(self):
        return self.pp.ip.get_phases_si(self.number)
    @property
    def elements(self):
        return self.pp.ip.get_elements_totals(self.number)
    @property
    def species(self, units='mmol'):
        return self.pp.ip.get_species_moles(self.number)
    @property
    def species_moles(self, units='mmol'):
        return self.pp.ip.get_species_moles(self.number)
    @property
    def species_molalities(self, units='mmol'):
        return self.pp.ip.get_species_molalities(self.number)
    @property
    def species_activities(self, units='mmol'):
        return self.pp.ip.get_species_activities(self.number)
    @property
    def masters_species(self):
        """ Returns a Phreeqc output like species table """
        return self.pp.ip.get_masters_species(self.number)

    # pretty printing
    def __str__(self):
        return "<PhreeqPython.Solution."+self.__class__.__name__ + " with number '" + self.number + "'>"



""" Phreeqpython module """

import os
import gzip
from .viphreeqc import VIPhreeqc
from .solution import Solution
from .gas import Gas
import warnings

class PhreeqPython(object):
    """ PhreeqPython Class to interact with the VIPHREEQC module """

    def __init__(self, database=None, from_file=None, debug=False):
        # Create VIPhreeqc Instance
        self.ip = VIPhreeqc()
        self.ip.debug = debug
        # Load Vitens.dat database. The VIPhreeqc module is unable to handle relative paths
        if not database:
            database_path = os.path.dirname(__file__) + "/database/vitens.dat"
        else:
            database_path = os.path.dirname(__file__) + "/database/"+database

        if not os.path.isfile(database_path):
            raise FileNotFoundError("Database file not found")

        self.ip.load_database(database_path)


        if from_file:
            dump = gzip.open(from_file,"rb")
            try:
                inputstr = dump.read().decode('utf-8') + "END"
                self.ip.run_string(inputstr)

                solutions = self.ip.get_solution_list()
                # force IPHREEQC to calculate all the solution properties
                for solution_number in solutions:
                    self.change_solution(solution_number,{'Na':0})

                self.solution_counter = solutions[-1]
                # precalculte all solutions
            finally:
                dump.close()

        else:
            self.buffer = False
            self.solution_counter = -1
            self.gas_counter = -1

    def add_gas(self, components=None, pressure=1.0, volume=1.0, fixed_pressure=True, fixed_volume=False, equilibrate_with=False):
        """ add a gas phase to the VIPhreeqc stack """

        if fixed_pressure and fixed_volume:
            raise ValueError("Cannot create a gas with a fixed pressure and a fixed volume!")

        if equilibrate_with is not False and not fixed_volume:
            raise ValueError("Only gas phases with a fixed_volume can be created in equilibrium with a solution")

        self.gas_counter += 1


        inputstr = "GAS_PHASE {}\n".format(self.gas_counter)
        if fixed_pressure:
            inputstr += "-fixed_pressure \n"
        if fixed_volume:
            inputstr += "-fixed_volume \n"

        inputstr += "-volume {}\n".format(volume)
        inputstr += "-pressure {}\n".format(pressure)

        for gas, pressure in components.items():
            inputstr += "{} {} \n".format(gas, pressure)

        if equilibrate_with:
            if isinstance(equilibrate_with, Solution):
                inputstr += "-equilibrate {}\n".format(equilibrate_with.number)
            else:
                inputstr += "-equilibrate {}\n".format(equilibrate_with)


        inputstr += "SAVE GAS_PHASE "+str(self.gas_counter) + "\n"
        inputstr += "END \n"

        self.ip.run_string(inputstr)

        return Gas(self, self.gas_counter)

    def add_solution_raw(self, composition=None):
        warnings.warn("add_solution_raw is deprecated, use add_solution and add_solution_simple instead", DeprecationWarning)
        return self.add_solution(composition)

    def add_solution(self, composition=None):
        """ add a solution to the VIPhreeqc Stack, allowing more control over the
        created solution """

        self.solution_counter += 1
        inputstr = "SOLUTION "+str(self.solution_counter) + "\n"
        if len(composition) > 0:
            for key, value in composition.items():
                inputstr += "  "+key+" "+str(value) + "\n"

        inputstr += "SAVE SOLUTION "+str(self.solution_counter) + "\n"
        inputstr += "END \n"

        self.ip.run_string(inputstr)

        return Solution(self, self.solution_counter)

    def add_solution_simple(self, composition=None, temperature=25):
        """ add a solution to the VIPhreeqc Stack and add all individual components 
        in a reaction step 
        """
        self.solution_counter += 1

        inputstr = "SOLUTION "+str(self.solution_counter) + "\n"
        inputstr += "-temp "+str(temperature) + "\n"
        if len(composition) > 0:
            inputstr += "REACTION 1 \n"
            for species, moles in composition.items():
                inputstr += species + " " + str(moles) + "\n"
            inputstr += "1 mmol \n"

        inputstr += "SAVE SOLUTION "+str(self.solution_counter) + "\n"
        inputstr += "END \n"

        self.ip.run_string(inputstr)

        return Solution(self, self.solution_counter)

    def add_master_species(self, element, master_species, alkalinity=0, gfw=1, egfw=""):
        """ add a master species to the VIPhreeqc Instance """
        inputstr = "SOLUTION_MASTER_SPECIES; {} {} {} {} {} \n".format(element, master_species, alkalinity, gfw, egfw)
        self.buffer = inputstr

    def add_species(self, reaction, log_k=0, delta_h=0, egfw=None):
        """ add a solution species to the VIPhreeqc Instance """
        inputstr = self.buffer if self.buffer else ""
        self.buffer = False
        inputstr +=  "SOLUTION_SPECIES \n {} \n".format(reaction)
        inputstr += "log_k {}".format(log_k)

        self.ip.run_string(inputstr)

    def change_solution(self, solution_number, elements, create_new=False):
        """ change solution composition by adding/removing elements """

        inputstr = "USE SOLUTION "+str(solution_number)+"\n"
        inputstr += "REACTION 1 \n"
        for element, change in elements.items():
            inputstr += element + " " + str(change) + "\n"
        inputstr += "1 mol \n"
        if create_new:
            self.solution_counter += 1
            solution_number = self.solution_counter

        inputstr += "SAVE SOLUTION "+str(solution_number) + "\n"
        inputstr += "END"

        self.ip.run_string(inputstr)
        return Solution(self, solution_number)

    def equalize_solution(self, solution_number, phases, to_si, in_phase=[10], with_element=[None]):
        """ saturate or desaturate (equalize) a solution with one or more phases """

        if not isinstance(phases, list):
            phases = [phases]
        if not isinstance(to_si, list):
            to_si = [to_si]
        if not isinstance(in_phase, list):
            in_phase = [in_phase]
        if not isinstance(with_element, list):
            with_element = [with_element]

        else:
            # fix default inputs
            if len(to_si) < len(phases):
                to_si.extend([0 for i in range(len(phases)-len(to_si))])

            if len(in_phase) < len(phases):
                in_phase.extend([10 for i in range(len(phases)-len(in_phase))])

            if len(with_element) < len(phases):
                with_element.extend([None for i in range(len(phases)-len(with_element))])

        inputstr = "USE SOLUTION "+str(solution_number)+"\n"
        inputstr += "EQUILIBRIUM PHASES 1 \n"

        for num in range(len(phases)):

            if with_element[num]:
                inputstr += phases[num] + " " + str(to_si[num]) + " " + with_element[num] + " " + str(in_phase[num]) + "\n"
            else:
                inputstr += phases[num] + " " + str(to_si[num]) + " " + str(in_phase[num]) + "\n"

        inputstr += "SAVE SOLUTION "+str(solution_number) + "\n"
        inputstr += "END"

        self.ip.run_string(inputstr)
        return Solution(self, solution_number)

    def mix_solutions(self, solutions):
        """ Create a mixture from two other solutions """
        # add a solution to the VIPhreeqc Stack
        self.solution_counter += 1
        # mix two or more solutions to obtain a new solution
        inputstr = "MIX 1 \n"
        # set of phreeqpython IDs
        pp_ids = set()
        for solution, fraction in solutions.items():
            if isinstance(solution, Solution):
                pp_ids.add(solution.pp.ip.id_)
                inputstr += str(solution.number) + " " + str(fraction) + "\n"
            else:
                inputstr += str(solution) + " " + str(fraction) + "\n"

        inputstr += "SAVE SOLUTION "+str(self.solution_counter) + "\n"
        inputstr += "END \n"

        if len(pp_ids) > 1:
            raise ValueError('Cannot Mix solutions belonging to seperate PhreeqPython instances!')

        self.ip.run_string(inputstr)

        return Solution(self, self.solution_counter)

    def interact_solution_gas(self, solution_number, gas_number):
        """ Interact solution with gas phase """
        inputstr = "USE SOLUTION " + str(solution_number) + "\n"
        inputstr += "USE GAS_PHASE " + str(gas_number) + "\n"
        inputstr += "SAVE GAS_PHASE " + str(gas_number) + "\n"
        inputstr += "SAVE SOLUTION " + str(solution_number) + "\n"
        inputstr += "END"
        self.ip.run_string(inputstr)


    def change_solution_temperature(self, solution_number, temperature):
        """ change temperature """
        inputstr = "USE SOLUTION " + str(solution_number) + "\n"
        inputstr += "REACTION_TEMPERATURE 1 \n"
        inputstr += str(temperature) + "\n"
        inputstr += "SAVE SOLUTION "+str(solution_number) + "\n"

        self.ip.run_string(inputstr)
        return Solution(self, self.solution_counter)

    def copy_solution(self, solution_number):
        """ Copy a solution to create a new one """
        # add a solution to the VIPhreeqc Stack
        self.solution_counter += 1
        # mix two or more solutions to obtain a new solution
        inputstr = "COPY SOLUTION " + str(solution_number) + " " + str(self.solution_counter) + "\n"
        inputstr += "END\n"

        self.ip.run_string(inputstr)

        return Solution(self, self.solution_counter)

    def copy_gas(self, gas_number):
        """ Copy a solution to create a new one """
        # add a solution to the VIPhreeqc Stack
        self.gas_counter += 1
        # mix two or more solutions to obtain a new solution
        inputstr = "COPY GAS_PHASE " + str(gas_number) + " " + str(self.gas_counter) + "\n"
        inputstr += "END\n"

        self.ip.run_string(inputstr)

        return Gas(self, self.gas_counter)

    def empty_solution(self):
        return self.add_solution({})

    def remove_solutions(self, solution_number_list):
        """ Remove solutions from VIPhreeqc memory """
        inputstr = "DELETE \n"
        inputstr += "-solution " + ' '.join(map(str, solution_number_list))
        self.ip.run_string(inputstr)

    def remove_gases(self, gas_number_list):
        """ Remove solutions from VIPhreeqc memory """
        inputstr = "DELETE \n"
        inputstr += "-gas_phase " + ' '.join(map(str, gas_number_list))
        self.ip.run_string(inputstr)

    def get_solution(self, number):
        return Solution(self, number)

    def dump_solutions(self, solution_number_list = None, filename='dump.gz'):
        """ Dump solutions to raw file for transmission to another VIPhreeqc instance """
        if not solution_number_list:
            solution_number_list = []

        inputstr = "DUMP \n"
        inputstr += "-file "+filename + "\n"
        inputstr += "-solution " + ' '.join(map(str, solution_number_list)) + "\n"
        inputstr += "END"

        self.ip.set_dump_string_on()
        self.ip.run_string(inputstr)

        dump = self.ip.get_dump_string()

        # write to file
        dumpfile = gzip.open(filename,'w')
        dumpfile.write(dump)
        dumpfile.close()

        self.ip.set_dump_string_off()

    def get_solution_list(self):
        return self.ip.get_solution_list()




from setuptools import setup
import os
import sys
import zipfile
import urllib

setup(name='phreeqpython',
      version='1.3.4',
      description='Vitens viphreeqc wrapper and utilities',
      url='https://github.com/Vitens/phreeqpython',
      author='Abel Heinsbroek',
      author_email='abel.heinsbroek@vitens.nl',
      license='Apache Licence 2.0',
      packages=['phreeqpython'],
      include_package_data=True,
      zip_safe=False,
      install_requires=['periodictable', 'scipy']
      )
